/* Zadatak 2A:

*/

#include <stdio.h>

/* ========================================================================= */

    /* Funkcija zaba iz (a). */

int zaba(int n, char A[], int korak)
{
    int i, broj_hrane = 0;

        /* Odmah provjerimo je li n djeljiv s korakom.
           Ako nije, zadnji skok ne moze biti na polje s indeksom n
           pa vratimo -1 (neuspjeh). */
    if (n % korak != 0) return -1;

        /* Petlja za skokove, svaki duljine korak.
           Nakon skoka, zaba je na polju s indeksom i. */
    for (i = korak; i <= n; i += korak) {

            /* Provjera za lose polje (neuspjeh). */
        if (A[i] == 'v') return -1;

            /* Provjera za polje s hranom. */
        if (A[i] == 'h') ++broj_hrane;
    }

        /* Ako prodjemo kroz cijelu petlju, prolaz je uspjesan.
           Vrati broj prikupljene hrane. */
    return broj_hrane;
}

/* ========================================================================= */

    /* Uociti da koraci u polju k strogo rastu, pa se najveci korak
       dobije kao zadnji korak za koji skupimo bar min hrane (ako ga ima).
       Za pocetne (male) korake mozemo dobiti neuspjeh, no za zadnji
       korak (k[m] = n) sigurno imamo uspjeh, jer je A[n] != 'v'. */

    /* Funkcija minhrana iz (b) - jedno binarno pretrazivanje.
       Jednim binarnim trazenjem preskacemo eventualne neuspjesne korake
       (na pocetku polja k) i trazimo najveci uspjesni korak s bar min hrane
       (ako ga ima). */

int minhrana(int n, int m, int min, char A[], int k[])
{
    int max_korak = -1;

        /* Trazimo u granicama indeksa [l, d], a s je srednji. */
    int l = 0, d = m, s;
        /* Trenutni korak = k[s] i pripadni broj hrane. */
    int korak, broj_hrane;

        /* Trazimo sve dok je l <= d, tj. skup preostalih indeksa
           za pretragu je neprazan. */
    while (l <= d) {
            /* Indeks na "pola" preostalog intervala.
               Uociti da za l = d ili l + 1 = d (tj. imamo 1 ili 2
               preostala indeksa u intervalu), dobijemo s = l. */
        s = (l + d) / 2;

            /* Pripadni korak = k[s]. */
        korak = k[s];
            /* Pripadni broj hrane - poziv funkcije iz (a). */
        broj_hrane = zaba(n, A, korak);

            /* Ako je prolaz neuspjesan, dizemo l na s + 1.
               Naime, za sve korake manje od ovog (ispred indeksa s),
               isto imamo neuspjeh. */
        if (broj_hrane == -1) {
            l = s + 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je prolaz uspjesan za trenutni korak. Ako je
               broj_hrane < min, onda to isto (zbog djeljivosti) vrijedi
               i za sve vece korake (iza indeksa s). Smanji d na s - 1. */
        if (broj_hrane < min) {
            d = s - 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je broj_hrane >= min. Za sve indekse ispred s
               (tj. za manje korake) dobivamo ili neuspjeh, ili jos veci
               broj hrane (>= min).
               Dakle, trazeni najveci korak je veci ili jednak od ovog. */

            /* Ako ZAPAMTIMO ovaj korak k[s] kao najveci (do sada)
               koji daje bar min hrane, onda nam on vise NE treba.
               Zato mozemo povecati l na s + 1 (sigurno raste za bar 1)
               i vise nema opasnosti od beskonacne petlje!
                 Uociti da u preostalom intervalu vise ne mora postojati
               korak koji daje bar min hrane, ali to nije bitno. */
        max_korak = korak;
        l = s + 1;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Glavni program za demo. */

int main(void)
{
        /* Polje A s 25 elemenata za demo. */
    int n = 24;
    char A[25] = {
        'l',    /* Pocetno polje zabe. */
        'l', 'l', 'v', 'v', 'l', 'h',  'l', 'l', 'h', 'l', 'l', 'h',
        'l', 'l', 'h', 'l', 'l', 'h',  'l', 'l', 'h', 'l', 'l', 'h'
      };

        /* Polje koraka k za demo. */
    int m = 4;
    int k[5] = { 1, 2, 6, 12, 24 };

        /* Ostale varijable za demo. */
    int korak, broj_hrane, min;

        /* Petlja za test funkcije iz (a) - razni koraci. */
    for (korak = 1; korak <= n; ++korak) {

            /* Poziv funkcije zaba iz (a). */
        broj_hrane = zaba(n, A, korak);

            /* Ispis rezultata za dani korak. */
        printf(" korak = %2d,  ", korak);
        if (broj_hrane == -1)
            printf("neuspjeh\n");
        else
            printf("broj hrane = %2d\n", broj_hrane);
    }

    printf("\n");

        /* Petlja za test funkcije iz (b), razni min. */
    for (min = 5; min >= 1; --min) {

            /* Poziv funkcije minhrana iz (b). */
        korak = minhrana(n, m, min, A, k);
        printf(" min = %2d,  ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);
    }

    return 0;
}
