/* Zadatak 2D:

*/

#include <stdio.h>

/* ========================================================================= */

    /* Funkcija istrazivac iz (a). */

int istrazivac(int n, char D[], int k)
{
    int i, broj_blaga = 0;

        /* Odmah provjerimo je li n djeljiv s korakom k.
           Ako nije, zadnji skok ne moze biti na polje s indeksom n
           pa vratimo -1 (neuspjeh). */
    if (n % k != 0) return -1;

        /* Petlja za skokove, svaki duljine k.
           Nakon skoka, istrazivac je na polju s indeksom i. */
    for (i = k; i <= n; i += k) {

            /* Provjera za lose polje (neuspjeh). */
        if (D[i] == 'o') return -1;

            /* Provjera za polje s blagom. */
        if (D[i] == 'b') ++broj_blaga;
    }

        /* Ako prodjemo kroz cijelu petlju, prolaz je uspjesan.
           Vrati broj prikupljenih blaga. */
    return broj_blaga;
}

/* ========================================================================= */

    /* Uociti da koraci (k-ovi) u polju k strogo rastu, pa se najveci k
       dobije kao zadnji k za koji skupimo bar min blaga (ako ga ima).
       Za pocetne (male) k-ove mozemo dobiti neuspjeh, no za zadnji
       k (k[m] = n) sigurno imamo uspjeh, jer je D[n] != 'o'. */

    /* Funkcija minblago iz (b) - jedno binarno pretrazivanje.
       Jednim binarnim trazenjem preskacemo eventualne neuspjesne k-ove
       (na pocetku polja k) i trazimo najveci uspjesni k s bar min blaga
       (ako ga ima). */

int minblago(int n, int m, int min, char D[], int k[])
{
    int max_korak = -1;

        /* Trazimo u granicama indeksa [l, d], a s je srednji. */
    int l = 0, d = m, s;
        /* Trenutni korak = k[s] i pripadni broj blaga. */
    int korak, broj_blaga;

        /* Trazimo sve dok je l <= d, tj. skup preostalih indeksa
           za pretragu je neprazan. */
    while (l <= d) {
            /* Indeks na "pola" preostalog intervala.
               Uociti da za l = d ili l + 1 = d (tj. imamo 1 ili 2
               preostala indeksa u intervalu), dobijemo s = l. */
        s = (l + d) / 2;

            /* Pripadni korak = k[s]. */
        korak = k[s];
            /* Pripadni broj blaga - poziv funkcije iz (a). */
        broj_blaga = istrazivac(n, D, korak);

            /* Ako je prolaz neuspjesan, dizemo l na s + 1.
               Naime, za sve korake manje od ovog (ispred indeksa s),
               isto imamo neuspjeh. */
        if (broj_blaga == -1) {
            l = s + 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je prolaz uspjesan za trenutni korak. Ako je
               broj_blaga < min, onda to isto (zbog djeljivosti) vrijedi
               i za sve vece korake (iza indeksa s). Smanji d na s - 1. */
        if (broj_blaga < min) {
            d = s - 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je broj_blaga >= min. Za sve indekse ispred s
               (tj. za manje korake) dobivamo ili neuspjeh, ili jos veci
               broj blaga (>= min).
               Dakle, trazeni najveci korak je veci ili jednak od ovog. */

            /* Ako ZAPAMTIMO ovaj korak k[s] kao najveci (do sada)
               koji daje bar min blaga, onda nam on vise NE treba.
               Zato mozemo povecati l na s + 1 (sigurno raste za bar 1)
               i vise nema opasnosti od beskonacne petlje!
                 Uociti da u preostalom intervalu vise ne mora postojati
               korak koji daje bar min blaga, ali to nije bitno. */
        max_korak = korak;
        l = s + 1;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Glavni program za demo. */

int main(void)
{
        /* Polje D s 25 elemenata za demo. */
    int n = 24;
    char D[25] = {
        'p',    /* Pocetno polje istrazivaca. */
        'p', 'p', 'o', 'o', 'p', 'b',  'p', 'p', 'b', 'p', 'p', 'b',
        'p', 'p', 'b', 'p', 'p', 'b',  'p', 'p', 'b', 'p', 'p', 'b'
      };

        /* Polje koraka k za demo. */
    int m = 4;
    int k[5] = { 1, 2, 6, 12, 24 };

        /* Ostale varijable za demo. */
    int korak, broj_blaga, min;

        /* Petlja za test funkcije iz (a) - razni koraci (k-ovi). */
    for (korak = 1; korak <= n; ++korak) {

            /* Poziv funkcije istrazivac iz (a). */
        broj_blaga = istrazivac(n, D, korak);

            /* Ispis rezultata za dani korak. */
        printf(" korak = %2d,  ", korak);
        if (broj_blaga == -1)
            printf("neuspjeh\n");
        else
            printf("broj blaga = %2d\n", broj_blaga);
    }

    printf("\n");

        /* Petlja za test funkcije iz (b), razni min. */
    for (min = 5; min >= 1; --min) {

            /* Poziv funkcije minblago iz (b). */
        korak = minblago(n, m, min, D, k);
        printf(" min = %2d,  ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);
    }

    return 0;
}
