/* Zadatak 2A:

Niz $A$ sastoji se od $n + 1$ znakova. Svaki od znakova na mjestima s indeksom
od $0$ do $n$ je jedan od znakova \texttt{'l'} (lopoč), \texttt{'v'} (voda)
ili \texttt{'h'} (lopoč s hranom).
Na mjestu $n$ se \textbf{ne} nalazi znak \texttt{'v'}.
Na početku niza (indeks~$0$) nalazi se žaba, a pripadni znak je \texttt{'l'}.
\begin{itemize}
  \item[(a)]
Napišite funkciju \texttt{zaba} koja prima prirodni broj $n$, niz $A$
duljine $n + 1$ (kao gore) te prirodni broj \textit{korak}. Žaba skače
unaprijed skokovima duljine \textit{korak}. Uspjehom ćemo smatrati ako žaba
na ovakav način može doći do polja $n$ (zadnji skok mora biti točno na polje
$n$), bez da pritom stane na polje s oznakom \texttt{'v'}.
Ukoliko žaba može uspješno proći kroz niz, funkcija treba vratiti broj
prikupljene hrane. U suprotnom, funkcija vraća $-1$.
  \item[(b)]
Napišite funkciju \texttt{minhrana} koja prima prirodne brojeve $n$,
$m$ i \textit{min}, niz $A$ (kao ranije) te niz "koraka" $k$ s $m + 1$
elemenata. Pretpostavljamo (ne treba provjeravati) da je niz $k$ strogo
rastući i vrijedi
\begin{displaymath}
	k_0 = 1, \quad k_m = n, \quad
	k_{l - 1} \hbox{ dijeli } k_l, \hbox{ za } l = 1, \ldots, m.
\end{displaymath}
Funkcija treba vratiti najveći "korak" $k_l$, za koji će žaba ostvariti
uspješan prolazak kroz niz te, pritom, prikupiti barem \textit{min} hrane,
ako takav korak postoji. Ako ga nema, treba vratiti $-1$.
\end{itemize}
Rješenje (b) dijela zadatka koje koristi više od
$2 + 2 \lfloor \log_2 (m + 1) \rfloor$
poziva funkcije iz (a) dijela nosi najviše $15$ bodova.

*/

#include <stdio.h>

/* ========================================================================= */

    /* Funkcija zaba iz (a). */

int zaba(int n, char A[], int korak)
{
    int i, broj_hrane = 0;

        /* Odmah provjerimo je li n djeljiv s korakom.
           Ako nije, zadnji skok ne moze biti na polje s indeksom n
           pa vratimo -1 (neuspjeh). */
    if (n % korak != 0) return -1;

        /* Petlja za skokove, svaki duljine korak.
           Nakon skoka, zaba je na polju s indeksom i. */
    for (i = korak; i <= n; i += korak) {

            /* Provjera za lose polje (neuspjeh). */
        if (A[i] == 'v') return -1;

            /* Provjera za polje s hranom. */
        if (A[i] == 'h') ++broj_hrane;
    }

        /* Ako prodjemo kroz cijelu petlju, prolaz je uspjesan.
           Vrati broj prikupljene hrane. */
    return broj_hrane;
}

/* ========================================================================= */

    /* Uociti da koraci u polju k strogo rastu, pa se najveci korak
       dobije kao zadnji korak za koji skupimo bar min hrane (ako ga ima).
       Za pocetne (male) korake mozemo dobiti neuspjeh, no za zadnji
       korak (k[m] = n) sigurno imamo uspjeh, jer je A[n] != 'v'. */

/* ========================================================================= */

    /* Funkcija minhrana iz (b) - sekvencijalno, uzlazno po l. */

int minhrana_uzlazno(int n, int m, int min, char A[], int k[])
{
    int l, korak, broj_hrane, max_korak = -1;

        /* Sekvencijalna petlja unaprijed po svim koracima u polju k.
           Koraci strogo rastu, trazimo zadnjeg "dobrog". */
    for (l = 0; l <= m; ++l) {
        korak = k[l];

            /* Poziv funkcije iz (a) za korak k[l]. */
        broj_hrane = zaba(n, A, korak);

            /* Provjera rezultata za taj korak. Ako je broj_hrane >= min,
               onda je sigurno uspjeh (jer min >= 1). Samo zapamtimo
               ovaj korak (veci je od prethodnog, ako ga je bilo). */
        if (broj_hrane >= min)
            max_korak = korak;

            /* Ubrzanje: Ako je prolaz uspjesan, ali je broj_hrane < min,
               onda mozemo prekinuti petlju. Zbog djeljivosti koraka,
               svi veci koraci su uspjesni, ali daju jos manje hrane. */
        else    /* Ovdje je broj_hrane < min. */
            if (broj_hrane >= 0) break;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Funkcija minhrana iz (b) - sekvencijalno, silazno po l. */

int minhrana_silazno(int n, int m, int min, char A[], int k[])
{
    int l, korak, broj_hrane, max_korak = -1;

        /* Sekvencijalna petlja unatrag po svim koracima u polju k.
           Koraci strogo padaju, trazimo prvog "dobrog". */
    for (l = m; l >= 0; --l) {
        korak = k[l];

            /* Poziv funkcije iz (a) za korak k[l]. */
        broj_hrane = zaba(n, A, korak);

            /* Provjera rezultata za taj korak. Ako je broj_hrane >= min,
               onda je sigurno uspjeh (jer min >= 1). Samo zapamtimo
               ovaj korak i odmah prekinemo petlju (svi buduci koraci su
               manji od ovog). */
        if (broj_hrane >= min) {
            max_korak = korak;
            break;
        }

            /* Ubrzanje: Ako naletimo na neuspjeh, isto mozemo prekinuti
               petlju. Zbog djeljivosti koraka, svi manji koraci su
               sigurno neuspjesni. */
        if (broj_hrane == -1) break;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Funkcija minhrana iz (b) - dva binarna pretrazivanja.
       Prvo binarno trazenje trazi najmanji korak koji je uspjesan,
       a drugo trazi najveci uspjesni korak s bar min hrane (ako ga ima),
       u bloku uspjesnih koraka (nije prazan, zbog zadnjeg k[m] = n). */

int minhrana_bin2(int n, int m, int min, char A[], int k[])
{
    int max_korak = -1;

        /* Trazimo u granicama indeksa [l, d], a s je srednji. */
    int l = 0, d = m, s;
        /* Trenutni korak = k[s] i pripadni broj hrane. */
    int korak, broj_hrane;

        /* Prvo binarno trazenje: trazimo najmanji indeks l takav da
           je prolaz uspjesan za korak = k[l]. Takav korak sigurno
           postoji, tj. vrijedi l <= m, jer za najveci korak k[m] = n
           sigurno dobivamo uspjesan prolaz (zbog A[n] != 'v').
           Dakle, na pocetku, za k[d] = k[m] imamo uspjesan prolaz. */

        /* Trazimo sve dok je l < d, tj. skup preostalih indeksa
           za pretragu ima bar dva elementa, s tim da desnu granicu d
           postavljamo tako da za k[d] imamo uspjesan prolaz. */
    while (l < d) {
            /* Indeks na "pola" preostalog intervala.
               Uociti da za l + 1 = d (tj. imamo tocno 2 preostala
               indeksa u intervalu), dobijemo s = l. */
        s = (l + d) / 2;

            /* Pripadni korak = k[s]. */
        korak = k[s];
            /* Pripadni broj hrane - poziv funkcije iz (a). */
        broj_hrane = zaba(n, A, korak);

            /* Ako je prolaz neuspjesan, dizemo l na s + 1.
               Naime, za sve korake manje od ovog (ispred indeksa s),
               isto imamo neuspjeh. U protivnom, spustamo d bas na s.
               Ovdje NEMA opasnosti od beskonacne petlje, jer l dizemo
               bar za 1, ili d spustamo bar za 1. */
        if (broj_hrane == -1)
            l = s + 1;
        else
            d = s;
    }
        /* Ovdje je l = d, i za korak k[l] sigurno imamo uspjesan prolaz.
           Trazeni najveci korak, ako postoji, ima indeks u intervalu
           [l, m]. Zato vratimo d na m. */
    d = m;

        /* Drugo binarno trazenje: trazimo najveci indeks s takav da
           (uspjesan) prolaz za korak = k[s] daje barem min hrane, ako
           takav indeks postoji u [l, d]. */

        /* Trazimo sve dok je l <= d, tj. skup preostalih indeksa
           za pretragu je neprazan. */
    while (l <= d) {
            /* Indeks na "pola" preostalog intervala.
               Uociti da za l = d ili l + 1 = d (tj. imamo 1 ili 2
               preostala indeksa u intervalu), dobijemo s = l. */
        s = (l + d) / 2;

            /* Pripadni korak = k[s]. */
        korak = k[s];
            /* Pripadni broj hrane - poziv funkcije iz (a).
               Prolaz je sigurno uspjesan, tj. broj_hrane >= 0. */
        broj_hrane = zaba(n, A, korak);

            /* Ako je broj_hrane < min, onda to isto (zbog djeljivosti)
               vrijedi i za sve vece korake (iza indeksa s). Smanji d
               na s - 1. */
        if (broj_hrane < min) {
            d = s - 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je broj_hrane >= min. Za sve indekse ispred s
               (tj. za manje korake) dobivamo jos veci broj hrane (>= min).
               Dakle, trazeni najveci korak je veci ili jednak od ovog. */

/* SPORIJA varijanta, linije koda imaju komentar na pocetku linije.
//
//          /* To znaci da mozemo povecati l na s (NE na s + 1).
//             Tu treba oprez, da izbjegnemo beskonacnu petlju!
//             Ako je l = d ili l + 1 = d, onda je s = l, pa se l NECE
//             promijeniti. Zato testiramo je li l < s, i tada zaista
//             povecamo l na s. U protivnom, imamo dodatne testove. */
//      if (l < s)
//          l = s;
//          continue;    /* Zato da ispod nema else, jer se teze cita. */
//      }
//
//          /* Ovdje je l = d ili l + 1 = d. Ako je l = d, onda smo nasli
//             najveci korak = k[l]. U protivnom, provjerimo broj hrane
//             za korak k[d] = k[l + 1]. */
//      max_korak = korak;
//      if (l < d) {
//              /* Pripadni korak = k[d]. */
//          korak = k[d];
//              /* Pripadni broj hrane - poziv funkcije iz (a). */
//          broj_hrane = zaba(n, A, korak);
//          if (broj_hrane >= min)
//             max_korak = korak;
//      }
//          /* Nasli smo najveci korak - prekini petlju.
//             Umjesto break, mozemo staviti l = d + 2 ili d = l - 2. */
//      break;

/* BRZA i JEDNOSTAVNIJA varijanta. */
            /* Ako ZAPAMTIMO ovaj korak k[s] kao najveci (do sada)
               koji daje bar min hrane, onda nam on vise NE treba.
               Zato mozemo povecati l na s + 1 (sigurno raste za bar 1)
               i vise nema opasnosti od beskonacne petlje!
                 Uociti da u preostalom intervalu vise ne mora postojati
               korak koji daje bar min hrane, ali to nije bitno. */
        max_korak = korak;
        l = s + 1;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Funkcija minhrana iz (b) - jedno binarno pretrazivanje.
       Jednim binarnim trazenjem preskacemo eventualne neuspjesne korake
       (na pocetku polja k) i trazimo najveci uspjesni korak s bar min hrane
       (ako ga ima). */

int minhrana(int n, int m, int min, char A[], int k[])
{
    int max_korak = -1;

        /* Trazimo u granicama indeksa [l, d], a s je srednji. */
    int l = 0, d = m, s;
        /* Trenutni korak = k[s] i pripadni broj hrane. */
    int korak, broj_hrane;

        /* Trazimo sve dok je l <= d, tj. skup preostalih indeksa
           za pretragu je neprazan. */
    while (l <= d) {
            /* Indeks na "pola" preostalog intervala.
               Uociti da za l = d ili l + 1 = d (tj. imamo 1 ili 2
               preostala indeksa u intervalu), dobijemo s = l. */
        s = (l + d) / 2;

            /* Pripadni korak = k[s]. */
        korak = k[s];
            /* Pripadni broj hrane - poziv funkcije iz (a). */
        broj_hrane = zaba(n, A, korak);

            /* Ako je prolaz neuspjesan, dizemo l na s + 1.
               Naime, za sve korake manje od ovog (ispred indeksa s),
               isto imamo neuspjeh. */
        if (broj_hrane == -1) {
            l = s + 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je prolaz uspjesan za trenutni korak. Ako je
               broj_hrane < min, onda to isto (zbog djeljivosti) vrijedi
               i za sve vece korake (iza indeksa s). Smanji d na s - 1. */
        if (broj_hrane < min) {
            d = s - 1;
            continue;    /* Zato da ispod nema else, jer se teze cita. */
        }

            /* Ovdje je broj_hrane >= min. Za sve indekse ispred s
               (tj. za manje korake) dobivamo ili neuspjeh, ili jos veci
               broj hrane (>= min).
               Dakle, trazeni najveci korak je veci ili jednak od ovog. */

/* SPORIJA varijanta, linije koda imaju komentar na pocetku linije.
   Nacin razmislanja i kod su isti kao u prethodnoj funkciji. */
//
//          /* To znaci da mozemo povecati l na s (NE na s + 1).
//             Tu treba oprez, da izbjegnemo beskonacnu petlju!
//             Ako je l = d ili l + 1 = d, onda je s = l, pa se l NECE
//             promijeniti. Zato testiramo je li l < s, i tada zaista
//             povecamo l na s. U protivnom, imamo dodatne testove. */
//      if (l < s)
//          l = s;
//          continue;    /* Zato da ispod nema else, jer se teze cita. */
//      }
//
//          /* Ovdje je l = d ili l + 1 = d. Ako je l = d, onda smo nasli
//             najveci korak = k[l]. U protivnom, provjerimo broj hrane
//             za korak k[d] = k[l + 1]. */
//      max_korak = korak;
//      if (l < d) {
//              /* Pripadni korak = k[d]. */
//          korak = k[d];
//              /* Pripadni broj hrane - poziv funkcije iz (a). */
//          broj_hrane = zaba(n, A, korak);
//          if (broj_hrane >= min)
//             max_korak = korak;
//      }
//          /* Nasli smo najveci korak - prekini petlju.
//             Umjesto break, mozemo staviti l = d + 2 ili d = l - 2. */
//      break;

/* BRZA i JEDNOSTAVNIJA varijanta. */
            /* Ako ZAPAMTIMO ovaj korak k[s] kao najveci (do sada)
               koji daje bar min hrane, onda nam on vise NE treba.
               Zato mozemo povecati l na s + 1 (sigurno raste za bar 1)
               i vise nema opasnosti od beskonacne petlje!
                 Uociti da u preostalom intervalu vise ne mora postojati
               korak koji daje bar min hrane, ali to nije bitno. */
        max_korak = korak;
        l = s + 1;
    }

    return max_korak;
}

/* ========================================================================= */

    /* Glavni program za demo. */

int main(void)
{
        /* Polje A s 25 elemenata za demo. */
    int n = 24;
    char A[25] = {
        'l',    /* Pocetno polje zabe. */
        'l', 'l', 'v', 'v', 'l', 'h',  'l', 'l', 'h', 'l', 'l', 'h',
        'l', 'l', 'h', 'l', 'l', 'h',  'l', 'l', 'h', 'l', 'l', 'h'
      };

        /* Polje koraka k za demo. */
    int m = 4;
    int k[5] = { 1, 2, 6, 12, 24 };

        /* Ostale varijable za demo. */
    int korak, broj_hrane, min;

        /* Petlja za test funkcije iz (a) - razni koraci. */
    for (korak = 1; korak <= n; ++korak) {

            /* Poziv funkcije zaba iz (a). */
        broj_hrane = zaba(n, A, korak);

            /* Ispis rezultata za dani korak. */
        printf(" korak = %2d,  ", korak);
        if (broj_hrane == -1)
            printf("neuspjeh\n");
        else
            printf("broj hrane = %2d\n", broj_hrane);
    }

    printf("\n");

        /* Petlja za test funkcije iz (b), razni min. */
    for (min = 5; min >= 1; --min) {
        printf("\n");

            /* Pozivi funkcije minhrana iz (b). */
        korak = minhrana_uzlazno(n, m, min, A, k);
        printf(" min = %2d,  [sekv uzlazno] ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);

        korak = minhrana_silazno(n, m, min, A, k);
        printf(" min = %2d,  [sekv silazno] ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);

        korak = minhrana_bin2(n, m, min, A, k);
        printf(" min = %2d,  [bin 2 puta]   ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);

        korak = minhrana(n, m, min, A, k);
        printf(" min = %2d,  [bin jednom]   ", min);
        if (korak == -1)
            printf("nema takvog koraka\n");
        else
            printf("najveci korak = %2d\n", korak);
    }

    return 0;
}
